# -*- config:utf-8 -*-

from mailparse.__version__ import __version__
from mailparse.decoder import EmailDecode
from mailparse.encoder import EmailEncode

__all__ = (
    '__version__',
    'EmailDecode',
    'EmailEncode'
)



# -*- config:utf-8 -*-

from email import message_from_bytes, message_from_string
from email.utils import parseaddr, parsedate_to_datetime, formataddr, getaddresses
from email.header import decode_header
from email import policy
from utils import EMAIL_HEADERS, NOT_HEADERS, SINGLE_HEADERS, normalize_keys

import re, datetime, base64


__all__ = ('EmailDecode',)

# Note This list is different between Decoder and Encoder. That's why it's not in utils.py
_ignore_headers = ('content-transfer-encoding', 'content-type', 'timestamp')


class EmailDecode(dict):
    def __init__(self, email):
        self['headers'] = {}

        # We try to find the charset
        self.default_charset = email.get_content_charset()

        if self.default_charset is None:
            # If we haven't found it, we parse the body to find the correct one
            for c in email.get_charsets():
                if c is not None:
                    self.default_charset = c
                    break

        if self.default_charset is None and 'subject' in email:
            # Finally, our last try is with the subject
            for r in decode_header(email.get('subject')):
                if r[1] is not None:
                    self.default_charset = r[1]
                    break

        if self.default_charset:
            self.default_charset = self.default_charset.upper().replace('HTTP-EQUIVCONTENT-TYPE', '')

        for k in list(set(email.keys())):
            k = k.lower()
            if k in _ignore_headers:
                continue

            try:
                if k in NOT_HEADERS:
                    self._headers(k, email.get_all(k))
                else:
                    self._headers(normalize_keys(k), email.get_all(k), subpart='headers')
            except AttributeError:
                continue

        date = self.get('date')
        if date is not None:
            # Date is (should be) in RFC 2822 format
            if isinstance(date, list):
                date = date[0]

            try:
                parsed = parsedate_to_datetime(date)
                """
                `.timestamp` handles TZ Infos
                if parsed.tzinfo is not None and str(parsed.tzinfo) != 'UTC':
                    # Timezone aware when needed!
                    parsed = parsed.astimezone(pytz.UTC)
                """

                self['timestamp'] = int(parsed.timestamp())
            except (ValueError, TypeError, AssertionError):
                # Saw a "Year out of range" once! With a year of ... 1899 ...
                self['timestamp'] = int(datetime.datetime.utcnow().timestamp())

        self._process_payload(email)

    def _headers(self, key, values, subpart=None):
        base = self
        if subpart:
            base = self[subpart]

        # When subpart, the key has been normalized, so not lowered
        key_lower = key.lower()
        if key in EMAIL_HEADERS:
            values = [formataddr(x) for x in getaddresses(values)]

        for value in values:
            if not isinstance(value, (bytes, str)):
                value = str(value)

            if key_lower in EMAIL_HEADERS:
                value = self._parse_email(value)
                if not value:
                    continue

                if value['name']:
                    value['name'] = self._clear(value['name'])
            else:
                value = self._clear(value)

            if value is None or value == '':
                continue

            if key_lower in SINGLE_HEADERS:
                base[key] = value
            else:
                if key not in base:
                    base[key] = []

                base[key].append(value)

    def _clear(self, value):
        results = decode_header(value)

        if len(results) == 1:
            value = self._decode_str(results[0][0], results[0][1])
        else:
            parts = []
            for r in results:
                parts.append(self._decode_str(r[0], r[1]))
            value = ' '.join(parts)

        return re.sub(r"[\ ]{1,}", ' ', value.strip()).replace('\r', '\n').replace('\n', '')

    def _decode_str(self, value, charset=None):
        """
        Decodes a given str from a specific charset to UTF-8
        """
        if value is None:
            return None

        if charset:
            charset = charset.upper().replace('HTTP-EQUIVCONTENT-TYPE', '')
        else:
            charset = 'UTF-8'

        # First thing, we translate to charset
        # If it fails, if default != charset, we translate via self.default_charset
        # if it fails, and default != utf-8 and charset != 8, we translate via utf-8 (default)

        # If it's a bytes, we keep it that way
        # Otherwise:
        if isinstance(value, str):
            # We convert the value into bytes
            value = self._encode_charset(value, charset)

        # Now we return a str:
        return self._decode_charset(value, charset)

    def _encode_charset(self, value, charset=None):
        try:
            return value.encode(charset)
        except (LookupError, UnicodeEncodeError):
            pass

        # Given charset failed, we try with the default, if different
        if self.default_charset != charset:
            try:
                return value.encode(self.default_charset)
            except (LookupError, UnicodeEncodeError):
                pass

        if charset != 'UTF-8' and self.default_charset != 'UTF-8':
            try:
                return value.encode('UTF-8')
            except (LookupError, UnicodeEncodeError):
                pass

        # Fall back to an easy value
        return value.encode('utf-8', errors='xmlcharrefreplace')

    def _decode_charset(self, value, charset=None):
        try:
            return value.decode(charset)
        except (LookupError, UnicodeDecodeError):
            pass

        # Given charset failed, we try with the default, if different
        if self.default_charset and self.default_charset != charset:
            try:
                return value.decode(self.default_charset)
            except (LookupError, UnicodeDecodeError):
                pass

        if charset != 'UTF-8' and self.default_charset != 'UTF-8':
            try:
                return value.decode('UTF-8')
            except (LookupError, UnicodeDecodeError):
                pass

        # Fall back to an easy value
        return value.decode('utf-8', errors='replace')  # errors='backslashreplace')

    def _parse_email(self, email):
        (name, email) = parseaddr(email)
        if email == '' or email.find('@') == -1:
            return None

        if name == '':
            name = None

        return {'name': name, 'email': email}

    def _merge_multiple_payload(self, payload, default_charset, decode=True):
        charset = payload.get_content_charset() or default_charset

        if payload.get_content_type().lower() in ('message/rfc822', 'message/delivery-status'):
            content = []
            for pl in payload.get_payload():
                content.append(self._decode_str(bytes(pl.get_body())))

            content = ''.join(content)

        elif payload.is_multipart():
            content = []
            for pl in payload.get_payload():
                content.append(self._merge_multiple_payload(pl, charset))

            content = '\r\n\r\n'.join(content)
        else:
            content = self._decode_str(payload.get_payload(decode=decode), charset)

        return content

    def _process_payload(self, payload, default_charset=None):
        charset = payload.get_content_charset() or default_charset

        if payload.is_multipart() and payload.get_content_maintype().lower() != 'message':
            for pl in payload.get_payload():
                self._process_payload(pl, charset)
        else:
            if payload.get_content_disposition() != 'attachment' and payload.get_content_type().lower() == 'text/plain':
                self['text'] = self._decode_str(payload.get_payload(decode=True).strip(), charset)
            elif payload.get_content_disposition() != 'attachment' and payload.get_content_type().lower() == 'text/html':
                self['html'] = self._decode_str(payload.get_payload(decode=True).strip(), charset)
            else:
                attachment = {
                    'type': payload.get_content_type(),
                    'name': self._decode_str(payload.get_filename(), charset)
                }

                if payload.get('content-description') is not None:
                    attachment['description'] = self._decode_str(str(payload.get('content-description')), charset)

                if str(payload.get('Content-Transfer-Encoding')) == 'base64':
                    attachment['content'] = self._merge_multiple_payload(payload, charset, decode=False)
                elif payload.get_content_maintype().lower() in ('text', 'message'):
                    content = self._merge_multiple_payload(payload, charset)

                    if attachment['type'] != 'message/rfc822':
                        attachment['content'] = self._decode_str(content, charset)
                    else:
                        attachment['content'] = self._decode_str(content, charset)
                else:
                    attachment['content'] = base64.b64encode(self._decode_str(payload.get_payload(decode=True), charset).encode('utf-8')).decode('utf-8')

                attachment['content'] = attachment['content'].strip()

                key = 'attachments'
                if payload.get_content_disposition() == 'inline' and payload.get_content_maintype().lower() == 'image' and payload.get('content-id'):
                    """
                    get_content_disposition() : returns the **lowercased** value of the Content-Disposition header if any, with either "attachment", "inline" or None
                    We only consider images to be inline, so any other format is an attachment
                    The image must have a content-id to be considered inline too
                    """
                    # We only consider images to be inline
                    key = 'inlines'
                    attachment['cid'] = str(payload.get('content-id'))[1:-1]

                if key not in self:
                    self[key] = []

                self[key].append(attachment)

    @classmethod
    def load(cls, data):
        if isinstance(data, bytes):
            mail = message_from_bytes(data, policy=policy.SMTPUTF8)
        elif isinstance(data, str):
            mail = message_from_string(data, policy=policy.SMTPUTF8)

        return EmailDecode(mail)

    @classmethod
    def open(cls, path):
        # We'll let Python fail in case the path is not correct
        with open(path, 'rb') as fp:
            return cls.load(fp.read())



# -*- config:utf-8 -*-

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.message import MIMEMessage
from email.mime.image import MIMEImage
from email.mime.audio import MIMEAudio
from email.mime.application import MIMEApplication
from email.message import Message
from email.encoders import encode_noop
from email.header import Header
from email.utils import quote, make_msgid, formataddr, formatdate
from utils import normalize_keys, EMAIL_HEADERS
from io import StringIO, BytesIO
from email.generator import Generator, BytesGenerator

__all__ = ('EmailEncode',)


# Note This list is different between Decoder and Encoder. That's why it's not in utils.py
_ignore_headers = ('received', 'content-transfer-encoding', 'content-type', 'mime-version', 'timestamp')


class EmailEncode(object):
    def __init__(self, obj):
        text_msg = None
        if obj.get('text') is not None:
            text_msg = MIMEText(self._add_blankline(obj.get('text')), 'plain', 'utf-8')

        html_msg = None
        if obj.get('html') is not None:
            html_msg = MIMEText(self._add_blankline(obj.get('html')), 'html', 'utf-8')

        if text_msg and html_msg:
            parent = MIMEMultipart('alternative')
            parent.attach(text_msg)
            parent.attach(html_msg)
        elif text_msg is not None:
            parent = text_msg
        elif html_msg is not None:
            parent = html_msg
        else:
            if not obj.get('inlines', []) and not obj.get('attachments', []):
                raise Exception('Missing body message (Either text or html).')

            if not obj.get('inlines') and len(obj.get('attachments', [])) == 1:
                item = obj.get('attachments')[0]
                parent = self._get_correct_mime(item.get('type'), self._add_blankline(item.get('content')), encoder=encode_noop)

            # Special case (yay) where the body is encrypted via PGP
            """
            @see https://pastebin.com/9NxMtWh6  for application/signature
            @see https://pastebin.com/Qv3MFFVL  for application/pgp-encrypted
            @see https://pastebin.com/4gF67Hmg  for application/pgp-encrypted (advanced)

            if len(obj.get('attachments', [])) == 1 and obj.get('attachments')[0].get('type') == 'application/pgp-encrypted':
                parent = MIMEMultipart('encrypted', _subparts={'protocol': 'application/pgp-encrypted'})

                parent.attach(text_msg)
                parent.attach(html_msg)
            """

        if obj.get('inlines', []):
            related = MIMEMultipart('related')
            related.attach(parent)
            for inline in obj.get('inlines'):
                item = self._get_correct_mime(inline.get('type'), self._add_blankline(inline.get('content')), encoder=encode_noop)
                related.attach(self._set_inline(item, inline))

            parent = related

        if obj.get('attachments', []):
            mixeds = []
            reports = []
            for attachment in obj.get('attachments'):
                maintype, subtype = attachment.get('type').lower().split('/')

                is_mixed = False
                if maintype == 'text':
                    attach = MIMEText(self._add_blankline(attachment.get('content')), subtype, 'utf-8')
                elif maintype == 'message':
                    msg = Message()
                    msg.set_payload(self._add_blankline(attachment.get('content')) + '\r\n', None)
                    attach = MIMEMessage(msg, subtype)
                else:
                    is_mixed = True
                    attach = MIMEApplication(self._add_blankline(attachment.get('content')), subtype, _encoder=encode_noop)
                    attach['Content-Transfer-Encoding'] = 'base64'

                attach = self._set_attachment(attach, attachment)
                if is_mixed:
                    mixeds.append(attach)
                else:
                    reports.append(attach)

            mixed = None
            if mixeds:
                mixed = MIMEMultipart('mixed')
                mixed.attach(parent)

            report = None
            if reports:
                report = MIMEMultipart('report')
                if not mixed:
                    report.attach(parent)

                for r in reports:
                    report.attach(r)

            if mixed is not None:
                if report:
                    mixed.attach(report)
                for m in mixeds:
                    mixed.attach(m)

                parent = mixed
            elif report is not None:
                parent = report

            if reports:
                parent.set_param('report-type', 'delivery-status')

        parent.set_param('charset', 'utf-8')

        headers = obj.get('headers')
        for key in obj.keys():
            if key in ('headers', 'html', 'text', 'inlines', 'attachments', 'bcc'):
                continue

            headers[normalize_keys(key)] = obj[key]

        for header in headers:
            if header.lower() in _ignore_headers:
                continue

            if not isinstance(headers[header], (dict, list)):
                parent[header] = self._encode_header(headers[header])
                continue

            items = headers[header]
            if isinstance(headers[header], dict):
                items = [headers[header]]

            for current in items:
                if isinstance(current, dict) and header.lower() in EMAIL_HEADERS:
                    if current.get('email') is None:
                        raise Exception('Invalid email provided. Field {0} is missing the email value!'.format(header.lower()))

                    parent[normalize_keys(header)] = formataddr((self._encode_header(current.get('name')), self._encode_header(current.get('email'))))
                else:
                    parent[normalize_keys(header)] = self._encode_header(current)

        if 'message-id' not in parent:
            msg_id = make_msgid()
            parent['Message-ID'] = '{0}@improvmx.com>'.format(msg_id.split('@')[0])

        if 'date' not in parent:
            parent['Date'] = formatdate()

        self.email = parent

    def _get_correct_mime(self, content_type, data, encoder=None):
        maintype, subtype = content_type.lower().split('/')

        if maintype == 'image':
            return MIMEImage(data, subtype, _encoder=encoder)
        elif maintype == 'audio':
            return MIMEAudio(data, subtype, _encoder=encoder)
        elif maintype == 'text':
            return MIMEText(data, subtype, 'utf-8')
        elif maintype == 'message':
            msg = Message()
            msg.set_payload(data + '\r\n', None)
            return MIMEMessage(msg, subtype)
        elif maintype == 'application':
            item = MIMEApplication(data, subtype, _encoder=encoder)
            item['Content-Transfer-Encoding'] = 'base64'
            return item

    def _set_inline(self, item, inline):
        if inline.get('name'):
            item.set_param('name', quote(inline.get('name')))
            item['Content-Disposition'] = 'inline; filename="{0}"'.format(quote(inline.get('name')))

        item['Content-Transfer-Encoding'] = 'base64'
        item['Content-ID'] = '<{0}>'.format(inline.get('cid'))
        item['X-Attachment-Id'] = inline.get('cid')
        return item

    def _set_attachment(self, item, attachment):
        item.set_type(attachment.get('type'))
        if attachment.get('name'):
            item.set_param('name', quote(attachment.get('name')))
            item['Content-Disposition'] = 'attachment; filename="{0}"'.format(quote(attachment.get('name')))

        if attachment.get('description'):
            item['Content-Description'] = attachment.get('description')

        return item

    def _encode_header(self, value):
        if value is None:
            return None

        try:
            value.encode('ascii')
            return value
        except UnicodeEncodeError:
            return Header(value, 'utf-8').encode()

    def __str__(self):
        fp = StringIO()
        g = Generator(fp, mangle_from_=False, maxheaderlen=60)
        g.flatten(self.email)
        return fp.getvalue()

    def __bytes__(self):
        fp = BytesIO()
        g = BytesGenerator(fp, mangle_from_=False, maxheaderlen=60)
        g.flatten(self.email)
        return fp.getvalue()

    def _add_blankline(self, payload):
        return '{0}\r\n'.format(payload.strip())



# -*- config:utf-8 -*-

"""Miscellaneous utilities."""

__all__ = (
    'EMAIL_HEADERS',
    'NOT_HEADERS',
    'SINGLE_HEADERS',
    'normalize_key'
)

NOT_HEADERS = ('subject', 'from', 'to', 'cc', 'bcc', 'return-path', 'date', 'sender', 'message-id')
SINGLE_HEADERS = ('subject', 'from', 'date', 'sender', 'message-id', 'mime-version', 'return-path', 'delivered-to', 'x-forwarding-service', 'feedback-id')
EMAIL_HEADERS = ('from', 'return-path', 'to', 'cc', 'bcc', 'delivered-to', 'sender', 'reply-to')

def normalize_keys(key):
    normalized = []
    for x in key.lower().split('-'):
        if x in ('arc', 'dkim', 'spf'):
            normalized.append(x.upper())
        else:
            normalized.append('{0}{1}'.format(x[0:1].upper(), x[1:]))

    return '-'.join(normalized)



# -*- config:utf-8 -*-

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from codecs import open
from os import path
import re

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'src/__version__.py'), encoding='utf-8') as fp:
    try:
        version = re.findall(
            r"^__version__ = \"([^']+)\"\r?$", fp.read(), re.M
        )[0]
    except IndexError:
        raise RuntimeError("Unable to determine version.")

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='mailparse',
    version=version,
    license='MIT',
    description='Encode raw emails into Python dict objects and build raw emails from Python dict.',
    long_description_content_type='text/markdown',
    long_description=long_description,
    url='https://github.com/cnicodeme/mailparse',
    author='Cyril Nicodeme',
    author_email='contact@cnicodeme.com',
    packages=find_packages(),
    platforms='any',

    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',

        'Topic :: Communications :: Email',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
        'Topic :: Text Processing',

        'License :: OSI Approved :: MIT License',

        "Operating System :: OS Independent",
        "Programming Language :: Python",

        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],

    keywords='mail email parse parser encode decode encoder decoder eml',
    project_urls={
        'Official Website': 'https://github.com/cnicodeme/mailparse',
        'Documentation': 'https://github.com/cnicodeme/mailparse',
        'Source': 'https://github.com/cnicodeme/mailparse',
    },
)
